#include <dos/dos.h>
#include <dos/dosasl.h>
#include <dos/dosextens.h>
#include <dos/exall.h>
#include <dos/rdargs.h>
#include <exec/memory.h>
#include <exec/types.h>
#include <utility/utility.h>

#include <proto/dos.h>
#include <proto/exec.h>
#include <proto/utility.h>

#define BUFFER_SIZE   100


/*****************************************************************************

    NAME */

        BOOL IsDosEntryA(

/*  SYNOPSIS */

        char  * Name,
        ULONG   Flags,
	struct DosLibrary *DOSBase,
	struct Library *UtilityBase)

/*  LOCATION */

/*  FUNCTION

        There is a need in file/directory processing where an application
        may need to determine whether a path is just a volume/device or
        assignment name.

    INPUTS

        Name  - The path to test.

        Flags - Any combination of the following:

                LDF_ASSIGNS
                LDF_DEVICES
                LDF_VOLUMES

    RESULT

        Boolean True or False.

    NOTES

        In future releases the buffer size will be set via a taglist.

    EXAMPLE

        BOOL Success;

        ...

        Success = IsDosEntryA("Work:", LDF_VOLUMES, DOSBase, UtilityBase)
        if (Success)
        {
          ...
        }  

    BUGS

    SEE ALSO

        <dos/dosextens.h>

    INTERNALS

    HISTORY

        27-Jul-1997     laguest     Initial inclusion into the AROS tree

******************************************************************************/
{
  struct DosList *DList;
  int             Position;
  BOOL            ReturnValue;
  char            Buffer[BUFFER_SIZE + 1];

  ReturnValue = FALSE;

  Position = SplitName(Name, ':', &Buffer[0], 0, BUFFER_SIZE + 1);
  if (Position != -1 && Name[Position] == NULL)
  {
    DList = AttemptLockDosList(Flags | LDF_READ);
    if (DList != NULL)
    {
      ReturnValue = FindDosEntry(DList, &Buffer[0], Flags) ? TRUE : FALSE;

      UnLockDosList(Flags | LDF_READ);
    }
  }

  return (ReturnValue);

} /* IsDosEntry */
