;/*
sc link debug=line idir=include: nostackcheck opt data=far nostartup mountnew.c
exit
    (C) 1995-2001 AROS - The Amiga Research OS
    $Id: mount.c,v 1.9 2017/04/24 14:47:37 piru Exp $

    Desc: Mount CLI command
    Lang: English
*/

#include <exec/devices.h>
#include <exec/io.h>
#include <exec/memory.h>
#include <exec/semaphores.h>
#include <exec/types.h>
#include <dos/dosextens.h>
#include <dos/exall.h>
#include <dos/filehandler.h>
#include <dos/rdargs.h>
#include <libraries/configvars.h>
#include <libraries/expansion.h>
#include <workbench/workbench.h>
#include <workbench/startup.h>
#include <intuition/intuition.h>
#include <devices/trackdisk.h>
#include <libraries/query.h>

#include <proto/exec.h>
#include <proto/dos.h>
#include <proto/utility.h>
#include <proto/icon.h>
#include <proto/expansion.h>
#include <proto/intuition.h>
#include <proto/query.h>

#ifndef __SASC
#else
typedef unsigned long	IPTR;
#endif
#include <resources/filesysres.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#ifndef __SASC
# define   DEBUG 1
# include  <aros/debug.h>
#endif

void *memclr(void*,size_t);

#define	DEBUG_PATCHDOSNODE(x)	if (DebugFlag) x;
#define DEBUG_MOUNT(x)		if (DebugFlag) x;
#define DEBUG_PREPAREFILE(x)	if (DebugFlag) x;
#define	DEBUG_MAKEDOSNODE(x)	if (DebugFlag) x;
#define	DEBUG_CHECK(x)		if (DebugFlag) x;


#define	MOUNTLIST	"DEVS:MountList"
#define	DOSDRIVERS	"DEVS:DOSDrivers/"
#define	STORAGEDRIVERS	"SYS:Storage/DOSDrivers/"
#define	PARAMSLENGTH	(sizeof(IPTR) * 4 + sizeof(struct DosEnvec))


enum
{
	ERR_SPECIAL = 5000,
	ERR_DEVICENOTFOUND,
	ERR_INVALIDKEYWORD
};

const char *SearchTable[]=
{
	"",
	"DEVS:DOSDrivers/",
	"MOSSYS:DEVS/DOSDrivers/",
	"SYS:Storage/DOSDrivers/",
	"MOSSYS:Storage/DOSDrivers/",
	NULL
};

/*
 * startup,control need to be handled differently.
 */

const UBYTE options[]=
	"HANDLER/K,"
	"EHANDLER=FILESYSTEM/K,"
	"DEVICE/K,"
	"UNIT/K,"
	"FLAGS/K,"
	"SECTORSIZE=BLOCKSIZE/K,"
	"SURFACES/K,"
	"SECTORSPERTRACK=BLOCKSPERTRACK/K,"
	"SECTORSPERBLOCK/K,"
	"RESERVED/K,"
	"PREALLOC/K,"
	"INTERLEAVE/K,"
	"LOWCYL/K,"
	"HIGHCYL/K,"
	"BUFFERS/K,"
	"BUFMEMTYPE/K,"
	"MAXTRANSFER/K,"
	"MASK/K,"
	"BOOTPRI/K,"
	"DOSTYPE/K,"
	"BAUD/K,"
	"CONTROL/K,"
	"STACKSIZE/K,"
	"PRIORITY/K,"
	"GLOBVEC/K,"
	"STARTUP/K,"
	"MOUNT=ACTIVATE/K,"
	"FORCELOAD/K";

struct Args
{
	char	*Handler;
	char	*Filesystem;
	char	*Device;
	char	*Unit;
	char	*Flags;

	char	*BlockSize;
	char	*Surfaces;
	char	*BlocksPerTrack;
	char	*SectorsPerBlock;
	char	*Reserved;
	char	*PreAlloc;
	char	*Interleave;
	char	*LowCyl;
	char	*HighCyl;
	char	*Buffers;
	char	*BufMemType;
	char	*MaxTransfer;
	char	*Mask;
	char	*BootPri;
	char	*DosType;
	char	*Baud;
	char	*Control;

	char	*StackSize;
	char	*Priority;
	char	*GlobalVec;
	char	*Startup;
	char	*Activate;
	char	*ForceLoad;
};

#include "mount_version.h"

static const char version[] = VERSTAG;

void displayerror(CONST_STRPTR fmt, IPTR *args);
ULONG	CheckDevice(char	*name);
void	InitParams(IPTR		*params);
LONG	readfile(STRPTR name, STRPTR *mem, LONG *size);
ULONG	readmountlist(IPTR	*params,
                      STRPTR	name,
                      char	*mountlist);
ULONG	readmountfile(IPTR	*params,
                      STRPTR	name);
void preparefile(STRPTR buf, LONG size);
LONG	parsemountfile(IPTR		*params,
                       STRPTR		buf,
                       LONG		size);
LONG	parsemountlist(IPTR		*params,
                       STRPTR		name,
                       STRPTR		buf,
                       LONG		size);
LONG mount(IPTR		*params,
           STRPTR	name);

struct ExecBase		*SysBase;
struct DosLibrary	*DOSBase;
struct Library		*UtilityBase;
struct Library		*ExpansionBase;
struct Process		*MyProcess;

ULONG			StartupValue;
char			*StartupString=NULL;
char			*ControlString=NULL;
char			*FlagsString=NULL;
char			*UnitString=NULL;
ULONG			StackSize;
ULONG			Priority;
ULONG			Activate;
int			GlobalVec;
ULONG			ForceLoad;
char			*HandlerString;
char			*DeviceString;
struct Args		flagargs;
BOOL			DebugFlag=FALSE;
BOOL			NoFSSM=FALSE;


int main(void)
{
struct WBStartup *_WBenchMsg = NULL;
struct
{
	STRPTR *device;
	STRPTR from;
	LONG   debug;
} args;
IPTR		*params;
LONG		error = RETURN_FAIL;
struct RDArgs	*rda;
char            dirname[512];

  SysBase = *((struct ExecBase**) 4);
  MyProcess = (struct Process*) FindTask(NULL);

  if (!MyProcess->pr_CLI)
  {
    WaitPort(&MyProcess->pr_MsgPort);
    _WBenchMsg = (struct WBStartup *) GetMsg(&MyProcess->pr_MsgPort);
  }

  if ((DOSBase = (struct DosLibrary *)OpenLibrary("dos.library",37))!=0)
  {
    if ((UtilityBase = OpenLibrary("utility.library",37)))
    {
      if ((ExpansionBase = OpenLibrary("expansion.library",37)))
      {
        if (!_WBenchMsg)
        {
          memclr(&args,sizeof(args));

          if ((rda = ReadArgs("DEVICE/M,FROM/K,DEBUG/S", (LONG *)&args, NULL)))
          {
            STRPTR	*MyDevPtr;
            int		len;

            error = 0;

            DebugFlag	=	args.debug != 0;

            MyDevPtr	=	args.device;
            if (MyDevPtr)
            {
              while (*MyDevPtr && error == 0)
              {
                DEBUG_MOUNT(Printf("Mount: Current DevName <%s>\n",
                                   (ULONG)*MyDevPtr));

                memclr(&flagargs,sizeof(struct Args));



                if ((params = AllocVec(PARAMSLENGTH,
                                       MEMF_PUBLIC | MEMF_CLEAR)))
                {
                  StackSize			=	8192;
                  Priority			=	5;
                  GlobalVec			=	-1;
                  HandlerString		=	NULL;
                  DeviceString		=	NULL;
                  StartupString		=	NULL;
                  NoFSSM		=	FALSE;

                  len = strlen(*MyDevPtr);
                  if ((*MyDevPtr)[len-1] == ':')
                  {
                    /* search for a devicename */
                    DEBUG_MOUNT(Printf("Mount: search for devname <%s>\n",
                                       (ULONG)*MyDevPtr));

                    strcpy(dirname,
                           *MyDevPtr);
                    dirname[len-1]		=	'\0';

                    #if 0
                    if ((error=CheckDevice(dirname))!=RETURN_OK)
                    {
                      IPTR array[] = {dirname};
                      //DEBUG_MOUNT(Printf("Mount: is already mounted..stop\n"));

                      displayerror("Device '%s:' is already mounted", array);
                    }
                    else
                    #endif
                    {
                      if (args.from)
                      {
                        if ((error=readmountlist(params,
                                                 dirname,
                                                 args.from))!=RETURN_OK)
                        {
                          DEBUG_MOUNT(Printf("Mount: failed with error %ld\n",
                                             error));
                        }
                      }
                      else
                      {
                        char	**SearchPtr;
                        ULONG	slen;

                        DEBUG_MOUNT(Printf("Mount: search device definition <%s>\n",
                                           (ULONG)*MyDevPtr));
                        for (SearchPtr=(char**) SearchTable;
                             *SearchPtr;
                             SearchPtr++)
                        {
                          if(SetSignal(0L,SIGBREAKF_CTRL_C) & SIGBREAKF_CTRL_C)
                          {
                            error = RETURN_FAIL;
                            SetIoErr(ERROR_BREAK);
                            break;
                          }

                          slen = strlen(*SearchPtr);
                          strcpy(dirname,
                                 *SearchPtr);
                          dirname[slen]		=	'\0';
                          strcat(dirname, *MyDevPtr);
                          dirname[slen+len-1]	=	'\0';
                          DEBUG_MOUNT(Printf("Mount: try File <%s>\n",
                                             (ULONG)dirname));

                          if ((error=readmountfile(params,
                                                   dirname))!=RETURN_OK)
                          {
                            DEBUG_MOUNT(Printf("Mount: failed with error %ld\n",
                                               error));
                            if (error == ERROR_OBJECT_EXISTS)
                            {
                              break;
                            }
                          }
                          else
                          {
                            break;
                          }
                        }
                        if (error!=RETURN_OK && error!=ERROR_OBJECT_EXISTS)
                        {
                          DEBUG_MOUNT(Printf("Mount: try from mountlist\n"));
                          dirname[0]	=	'\0';
                          strcat(dirname, *MyDevPtr);
                          dirname[len-1]	=	'\0';
                          if ((error=readmountlist(params,
                                                   dirname,
                                                   MOUNTLIST))!=RETURN_OK)
                          {
                            DEBUG_MOUNT(Printf("Mount: failed with error %ld\n",
                                               error));
                          }
                        }
                      }
                    }
                  }
                  else
                  {
                    /* search for a filename */

                    LONG err;

                    UBYTE stack_ap[sizeof(struct AnchorPath) + 3];
                    struct AnchorPath	*MyAp = (struct AnchorPath *) (((IPTR) stack_ap + 3) & ~3);

                    DEBUG_MOUNT(Printf("Mount: search for mountfile <%s>\n",
                                       (ULONG)*MyDevPtr));

                    memclr(MyAp,sizeof(struct AnchorPath));

                    dirname[0]	=	'\0';
                    for (err = MatchFirst(*MyDevPtr,MyAp);
                         err == 0;
                         err = MatchNext(MyAp))
                    {
                      if (MyAp->ap_Flags & APF_DirChanged)
                      {
                        DEBUG_MOUNT(Printf("Mount: Changed directories...\n"));
                      }

                      if (NameFromLock(MyAp->ap_Current->an_Lock,
                                       dirname,
                                       sizeof(dirname)) == FALSE)
                      {
                        PrintFault(IoErr(),"Mount");
                        break;
                      }
                      if (AddPart(dirname,
                                  &(MyAp->ap_Info.fib_FileName[0]),
                                  sizeof(dirname)) == FALSE)
                      {
                        PrintFault(IoErr(),"Mount");
                        break;
                      }
                      if (MyAp->ap_Info.fib_DirEntryType > 0)
                      {
                        if (MyAp->ap_Flags & APF_DIDDIR)
                        {
                          DEBUG_MOUNT(Printf("Mount: Ascending from directory %s\n",
                                        (ULONG)dirname));
                        }
#if 0
                        else
                        if (MyArgs.All)
                        {
                          DEBUG_MOUNT(Printf("Mount: Entering directory %s\n",dirname));
                          /* make it enter the directory */
                          MyAp->ap_Flags   |=      APF_DODIR;
                        }
#endif 
                        else
                        {
                          DEBUG_MOUNT(Printf("Mount: The next dir is  ... %s\n", (ULONG)dirname));
                        }
                        /* clear the completed directory flag */
                        MyAp->ap_Flags     &=      ~APF_DIDDIR;

                      }
                      else
                      {
                        /* Here is code for handling each particular file */

                        DEBUG_MOUNT(Printf("Mount: try File <%s>\n",
                                           (ULONG)dirname));

                        if ((error=readmountfile(params,
                                                 dirname))!=RETURN_OK)
                        {
                          DEBUG_MOUNT(Printf("Mount: failed with error %ld\n",
                                             error));
                          #if 0
                          if (error == ERROR_OBJECT_EXISTS)
                          {
                            break;
                          }
                          #endif
                        }
                      }
                    }
                    /* This absolutely, positively must be called, all of the time. */
                    MatchEnd(MyAp);

                    if (err == ERROR_NO_MORE_ENTRIES)
                    {
                      SetIoErr(0);
                    }
                    else
                    {
                      /* if it was real error promote it - Piru */
                      error = err;
                    }
                  }
                  FreeVec(params);
                }
                else
                {
                  error = ERROR_NO_FREE_STORE;
                  break;
                }
                MyDevPtr++;
              }
            }
            FreeArgs(rda);
          } /* if (rda != NULL) */
          else
          {
            error = IoErr();
          }

          if (error && error != ERROR_NO_MORE_ENTRIES && error < ERR_SPECIAL)
          {
            if (error != ERROR_OBJECT_EXISTS)
            {
              PrintFault(error,"Mount");
            }

            error = RETURN_FAIL;
          }
          else
          {
            error = error < ERR_SPECIAL ? RETURN_OK : RETURN_FAIL;
          }

        }
        else
        {
          /* wb startup */
          if (_WBenchMsg->sm_NumArgs >= 2)
          {
            if ((params = AllocVec(PARAMSLENGTH,
                                   MEMF_PUBLIC | MEMF_CLEAR)))
            {
              int i;

              for (i = 1; i < _WBenchMsg->sm_NumArgs; i++)
              {
                BPTR olddir;

                DEBUG_MOUNT(Printf("Mount: try File <%s>\n",
                                   (ULONG) _WBenchMsg->sm_ArgList[i].wa_Name));

                olddir = CurrentDir(_WBenchMsg->sm_ArgList[i].wa_Lock);

                if ((error=readmountfile(params,
                                         _WBenchMsg->sm_ArgList[i].wa_Name))!=RETURN_OK)
                {
                  DEBUG_MOUNT(Printf("Mount: failed with error %ld\n",
                                     error));
                }

                (void) CurrentDir(olddir);
              }

              FreeVec(params);
            }
            else
            {
              error = ERROR_NO_FREE_STORE;
            }
          }
          
        }

        CloseLibrary((struct Library *)ExpansionBase);
      }
      CloseLibrary((struct Library *)UtilityBase);
    }
    CloseLibrary((struct Library *)DOSBase);
  }

  if (_WBenchMsg)
  {
    Forbid();
    ReplyMsg(&_WBenchMsg->sm_Message);
  }

  return error;
}

/************************************************************************************************/
/************************************************************************************************/
void displayerror(CONST_STRPTR fmt, IPTR *args)
{
  if (MyProcess->pr_CLI)
  {
    PutStr("ERROR: ");
    VPrintf(fmt, args);
    PutStr("\n");
  }
  else
  {
    struct Library *IntuitionBase;

    IntuitionBase = OpenLibrary("intuition.library", 37);
    if (IntuitionBase)
    {
      struct EasyStruct es;

      es.es_StructSize   = sizeof(es);
      es.es_Flags        = 0;
      es.es_Title        = "Mount Failure";
      es.es_TextFormat   = (STRPTR) fmt;
      es.es_GadgetFormat = "OK";
      (void) EasyRequestArgs(NULL, &es, NULL, args);

      CloseLibrary(IntuitionBase);
    }
  }
}

/************************************************************************************************/
/************************************************************************************************/
ULONG	CheckDevice(char	*name)
{
struct DosList	*dl;
ULONG		Status;

  DEBUG_CHECK(Printf("CheckDevice: <%s>\n",
                     name));

  dl = LockDosList(LDF_DEVICES | LDF_VOLUMES | LDF_ASSIGNS | LDF_READ);
  if ((dl = FindDosEntry(dl,name,LDF_DEVICES | LDF_VOLUMES | LDF_ASSIGNS)))
  {
    Status	=	ERROR_OBJECT_EXISTS;
  }
  else
  {
    Status	=	0;
  }
  UnLockDosList(LDF_DEVICES | LDF_VOLUMES | LDF_ASSIGNS | LDF_READ);

  DEBUG_CHECK(Printf("CheckDevice: object %sexists\n", Status ? "" : "doesn't "));

  return Status;
}

/************************************************************************************************/
/************************************************************************************************/

void	InitParams(IPTR		*params)
{
struct DosEnvec *vec;

  memclr(params,
         PARAMSLENGTH);

  vec				=(struct DosEnvec *)&params[4];

  vec->de_TableSize		=	DE_BOOTBLOCKS;
  vec->de_SizeBlock		=	512 >> 2;
  vec->de_Surfaces		=	2;
  vec->de_SectorPerBlock	=	1;
  vec->de_BlocksPerTrack	=	11;
  vec->de_Reserved		=	2;

/* memclr above
  vec->de_SecOrg		=	0;
  vec->de_BootBlocks		=	0;
  vec->de_BootPri		=	0;
  vec->de_PreAlloc		=	0;
  vec->de_Interleave		=	0;
  vec->de_LowCyl		=	0;
*/

  vec->de_HighCyl		=	79;
  vec->de_NumBuffers		=	20;
  vec->de_BufMemType		=	1;
  vec->de_Baud			=	1200; // 9600;
  vec->de_MaxTransfer		=	0x7fffffff; //0x00ffffff;
  vec->de_Mask			=	0xfffffffe; // ~0ul;
  vec->de_DosType		=	ID_DOS_DISK;

  memclr(&flagargs,sizeof(struct Args));

  StackSize			=	8192;
  Priority			=	5;
  GlobalVec			=	-1;
  HandlerString			=	NULL;
  DeviceString			=	NULL;
  StartupString			=	NULL;
  NoFSSM			=	FALSE;
}

void	FreeStuff(void)
{
  if (UnitString)
  {
    FreeVec(UnitString);
    UnitString		=	NULL;
  }
  if (FlagsString)
  {
    FreeVec(FlagsString);
    FlagsString		=	NULL;
  }
  if (ControlString)
  {
    FreeVec(ControlString);
    ControlString	=	NULL;
  }
  if (HandlerString)
  {
    FreeVec(HandlerString);
    HandlerString	=	NULL;
  }
  if (DeviceString)
  {
    FreeVec(DeviceString);
    DeviceString	=	NULL;
  }
  if (StartupString)
  {
    FreeVec(StartupString);
    StartupString	=	NULL;
  }
}

/************************************************************************************************/
/************************************************************************************************/




ULONG GetValueFlag(CONST_STRPTR buf, int *flag)
{
  ULONG	Value = 0;
  ULONG	Sign = 1;
  ULONG base = 10;

  if (*buf == '-')
  {
    Sign = -1;
    buf++;
  }
  #if 0
  else if (*buf == '+')
  {
    buf++;
  }
  #endif

  if (*buf == '0' && ToUpper(buf[1]) == 'X')
  {
    base = 16;
    buf += 2;
  }

  if (*buf == '\0')
  {
    if (flag) *flag = FALSE;
    return 0;
  }

  for (;;)
  {
    UBYTE ch;

    ch = *buf++;
    if (ch == '\0')
    {
      break;
    }
    ch = ToUpper((ULONG) ch);

    if (isdigit(ch))
    {
      Value *= base;
      Value += ch - '0';
    }
    else if (base == 16 && ch >= 'A' && ch <= 'F')
    {
      Value *= base;
      Value += ch - ('A' - 10);
    }
    else
    {
      if (flag) *flag = FALSE;
      return 0;
    }
  }

  if (flag) *flag = TRUE;
  return Value * Sign;
}

ULONG GetValue(CONST_STRPTR buf)
{
  return GetValueFlag(buf, NULL);
}


/************************************************************************************************/
/************************************************************************************************/

ULONG	ReadMountArgs(IPTR		*params,
                      struct RDArgs	*rda)
{
struct DosEnvec *vec;
struct Args	args;
struct RDArgs	*MyRDA;
ULONG		result=RETURN_OK;
int		i;

	DEBUG_MOUNT(Printf("ReadMountArgs:\n%s\n\n",(ULONG)&rda->RDA_Source.CS_Buffer[rda->RDA_Source.CS_CurChr]));

	memclr(&args,sizeof(struct Args));

	if (!(MyRDA=ReadArgs((STRPTR)options, (LONG*)&args, rda)))
	{
		DEBUG_MOUNT(Printf("ReadMountArgs: ReadArgs failed\n"));
		DEBUG_MOUNT(PrintFault(IoErr(),"ReadMountArgs"));
		//return (ULONG) IoErr();
		return  ERR_INVALIDKEYWORD;
	}

        vec = (struct DosEnvec *)&params[4];

	if (args.BlockSize)
	{
		vec->de_SizeBlock	=	GetValue(args.BlockSize) >> 2;
	}
	if (args.Surfaces)
	{
		vec->de_Surfaces	=	GetValue(args.Surfaces);
	}
	if (args.SectorsPerBlock)
	{
		vec->de_SectorPerBlock	=	GetValue(args.SectorsPerBlock);
	}
	if (args.BlocksPerTrack)
	{
		vec->de_BlocksPerTrack	=	GetValue(args.BlocksPerTrack);
	}
	if (args.Reserved)
	{
		vec->de_Reserved	=	GetValue(args.Reserved);
	}
	if (args.PreAlloc)
	{
		vec->de_PreAlloc	=	GetValue(args.PreAlloc);
	}
	if (args.Interleave)
	{
		vec->de_Interleave	=	GetValue(args.Interleave);
	}
	if (args.LowCyl)
	{
		vec->de_LowCyl		=	GetValue(args.LowCyl);
	}
	if (args.HighCyl)
	{
		vec->de_HighCyl		=	GetValue(args.HighCyl);
	}
	if (args.Buffers)
	{
		vec->de_NumBuffers	=	GetValue(args.Buffers);
	}
	if (args.BufMemType)
	{
		vec->de_BufMemType	=	GetValue(args.BufMemType);
	}
	if (args.BootPri)
	{
		vec->de_BootPri		=	GetValue(args.BootPri);
	}
	if (args.Baud)
	{
		vec->de_Baud		=	GetValue(args.Baud);
	}
	if (args.MaxTransfer)
	{
		vec->de_MaxTransfer	=	GetValue(args.MaxTransfer);
	}
	if (args.Mask)
	{
		vec->de_Mask		=	GetValue(args.Mask);
	}

	for (i=0;i<sizeof(args) / sizeof(IPTR);i++)
	{
		if (((ULONG*)&args)[i])
		{
			((ULONG*) &flagargs)[i]	=	TRUE;
		}
	}

	if (args.DosType)
	{
		vec->de_DosType		= (IPTR) GetValue(args.DosType);
	}

	if (args.GlobalVec)
	{
		GlobalVec		=	GetValue(args.GlobalVec);
	}

	if (args.BootPri)
	{
		vec->de_BootPri		=	GetValue(args.BootPri);
	}

	if (args.Startup)
	{
		char *String;
		int flag;

		DEBUG_MOUNT(Printf("ReadMountArgs: Startup <%s>\n",args.Startup));

		String	=	args.Startup;

		StartupValue = GetValueFlag(String, &flag);

		if (!flag)
		{
			int len;

			len = strlen(String);

			DEBUG_MOUNT(Printf("ReadMountArgs: len %ld\n",len));

			if (StartupString)
			{
				FreeVec(StartupString);
			}
			if ((StartupString = AllocVec(len+1,MEMF_PUBLIC)))
			{
				strcpy(StartupString,String);
			}
			else
			{
				result	=	ERROR_NO_FREE_STORE;
				goto error;
			}
		}
	}
	
	if (args.Control)
	{
		int len;
		DEBUG_MOUNT(Printf("ReadMountArgs: Control <%s>\n",args.Control));
		if (ControlString)
		{
			FreeVec(ControlString);
			ControlString	=	NULL;
		}
		len	=	strlen(args.Control);
		if (len < 0x100)
		{
			if ((ControlString=AllocVec(len+2,MEMF_PUBLIC)))
			{
				strcpy(&ControlString[1],args.Control);
				ControlString[0]	=	len;
				vec->de_Control		=	MKBADDR((char*) ControlString);
			}
			else
			{
				DEBUG_MOUNT(Printf("ReadMountArgs: Can't alloc Control string\n"));
				result	=	ERROR_NO_FREE_STORE;
				goto error;
			}
		}
		else
		{
			//Printf("Parse: Control uses a too large string..max 255 chars\n");
			result	=	ERROR_LINE_TOO_LONG;
			PrintFault(result, "Control");
			goto error;
		}
	}

	if (args.StackSize)
	{
		StackSize		=	GetValue(args.StackSize);
	}

	if (args.Priority)
	{
		Priority		=	GetValue(args.Priority);
	}

	if (args.ForceLoad)
	{
		ForceLoad		=(ULONG) GetValue(args.ForceLoad);
	}

	if (args.Activate)
	{
		Activate		=	GetValue(args.Activate);
	}

	if (args.Handler || args.Filesystem)
	{
		CONST_STRPTR str;
		int len;

		str = args.Handler;
		NoFSSM = str != NULL;
		if (!str)
		{
			str = args.Filesystem;
		}

		DEBUG_MOUNT(Printf("ReadMountArgs: %s <%s>\n", NoFSSM ? "Handler" : "FileSystem/EHandler", str));
		len			=	strlen(str);
		if (HandlerString)
		{
			FreeVec(HandlerString);
		}
		if ((HandlerString = AllocVec(len+2,MEMF_PUBLIC)))
		{
			strcpy(&HandlerString[1],str);
			HandlerString[0]	=	len;
		}
	}

	if (args.Device)
	{
		int len;

		DEBUG_MOUNT(Printf("ReadMountArgs: Device <%s>\n",args.Device));

		len		=	strlen(args.Device);

		//Printf("found args.Device\n");

		if (DeviceString)
		{
			FreeVec(DeviceString);
		}
		if ((DeviceString = AllocVec(len+1,MEMF_PUBLIC)))
		{
			//Printf("copying...\n");

			strcpy(DeviceString,args.Device);
		}
	}

	if (args.Unit)
	{
		char *String;
		int flag;

		DEBUG_MOUNT(Printf("ReadMountArgs: Unit <%s>\n",args.Unit));

		String = args.Unit;

		if (UnitString)
		{
			FreeVec(UnitString);
		}
		params[2] = GetValueFlag(String, &flag);
		if (flag)
		{
			DEBUG_MOUNT(Printf("ReadMountArgs: Unit Value %ld\n",params[2]));
		}
		else
		{
			int len;

			len = strlen(String);

			DEBUG_MOUNT(Printf("ReadMountArgs: len %ld\n",len));

			if ((UnitString = AllocVec(len+1,MEMF_PUBLIC)))
			{
				strcpy(UnitString,String);
				params[2]	=(ULONG) UnitString;
				DEBUG_MOUNT(Printf("ReadMountArgs: Unit String <%s>\n",params[2]));
			}
			else
			{
				result		=	ERROR_NO_FREE_STORE;
				goto error;
			}
		}
	}
	if (args.Flags)
	{
		char *String;
		int flag;

		DEBUG_MOUNT(Printf("ReadMountArgs: Flags <%s>\n",args.Flags));

		String	=	args.Flags;

		if (FlagsString)
		{
			FreeVec(FlagsString);
		}
		params[3] = GetValueFlag(String, &flag);
		if (flag)
		{
			DEBUG_MOUNT(Printf("ReadMountArgs: Flag Value %ld\n",params[3]));
		}
		else
		{
			int len;

			len = strlen(String);

			DEBUG_MOUNT(Printf("ReadMountArgs: len %ld\n",len));

			if ((FlagsString = AllocVec(len+1,MEMF_PUBLIC)))
			{
				strcpy(FlagsString,String);
				params[3]	=(ULONG) FlagsString;
				DEBUG_MOUNT(Printf("ReadMountArgs: Flags String <%s>\n",params[3]));
			}
			else
			{
				result		=	ERROR_NO_FREE_STORE;
				goto error;
			}
		}
	}
error:
	FreeArgs(MyRDA);

	return result;
}

/************************************************************************************************/
/************************************************************************************************/

ULONG	readmountlist(IPTR	*params,
                      STRPTR	name,
                      char	*mountlist)
{
STRPTR	MountListBuf;
LONG	MountListBufSize;
ULONG	error;

  DEBUG_MOUNT(Printf("ReadMountList: find <%s> in mountlist <%s>\n",
                     name,
                     mountlist));

  error = readfile(mountlist,
                   &MountListBuf,
                   &MountListBufSize);
  if (error==RETURN_OK)
  {
    preparefile(MountListBuf,
                MountListBufSize);


    InitParams(params);

    if ((error=parsemountlist(params,
                              name,
                              MountListBuf,
                              MountListBufSize))==RETURN_OK)
    {
      if ((error = mount(params,name))!=RETURN_OK)
      {
        DEBUG_MOUNT(Printf("ReadMountList: mount failed error %ld\n",
                           error));

        if (error == ERROR_OBJECT_EXISTS)
        {
          IPTR array[] = {name};
          displayerror("Device '%s:' is already mounted", array);
        }
      }
    }
#if 1
    else
    {
      switch (error)
      {
        case ERR_DEVICENOTFOUND:
          {
          IPTR array[] = {name, mountlist};
          displayerror("Device '%s:' not found in file '%s'", array);
          }
          break;
        case ERR_INVALIDKEYWORD:
          {
          IPTR array[] = {name};
          displayerror("Invalid keyword in file '%s'", array);
          }
          break;
      }
    }
#endif

    FreeStuff();
    FreeVec(MountListBuf);
  }
  return error;
}

/************************************************************************************************/
/************************************************************************************************/

ULONG	readmountfile(IPTR	*params,
                      STRPTR	filename)
{
struct Library		*IconBase;
struct DiskObject	*diskobj;
char			**myargv;
STRPTR			MountListBuf;
LONG			MountListBufSize;
struct RDArgs		rda;
ULONG			error = RETURN_FAIL;
char			*nameptr;
int			toollen;
BOOL			mountinfo=FALSE;
char			name[256+1];

  DEBUG_MOUNT(Printf("ReadMountFile: <%s>\n", (ULONG)filename));

  {
    struct Process *me = (APTR) FindTask(NULL);
    APTR oldwinptr;
    BPTR lock;

    name[0] = '\0';

    oldwinptr = me->pr_WindowPtr;
    me->pr_WindowPtr = (APTR) -1;
    lock = Lock(filename, MODE_OLDFILE);
    if (lock)
    {
      /* reusing name as fib, just to annoy certain guy ;-) */
      struct FileInfoBlock *fib = (APTR) ((((IPTR) name) + 3) & ~3);
      if (Examine(lock, fib))
      {
        nameptr = fib->fib_FileName;
        memmove(name, nameptr, strlen(nameptr) + 1);
      }
      UnLock(lock);
    }
    me->pr_WindowPtr = oldwinptr;

    if (name[0] == '\0')
    {
      nameptr		=	FilePart(filename);
      strcpy(name,
             nameptr);
    }
  }

  DEBUG_MOUNT(Printf("ReadMountFile: mount <%s>\n", (ULONG)name));

  #if 0
  if ((error=CheckDevice(name))!=RETURN_OK)
  {
    IPTR array[] = {name};
    displayerror("Device '%s:' is already mounted", array);

    /* Ignore error here, like the original - Piru */
    error = 0;

    return error;
  }
  #endif

  InitParams(params);

  DEBUG_MOUNT(Printf("ReadMountFile: readfile\n"));

  error = readfile(filename,
                   &MountListBuf,
                   &MountListBufSize);
  if (error==RETURN_OK)
  {
    DEBUG_MOUNT(Printf("ReadMountFile: preparsefile\n"));
    preparefile(MountListBuf, MountListBufSize);


    DEBUG_MOUNT(Printf("ReadMountFile: parsemountfile\n"));
    if ((error=parsemountfile(params,
                              MountListBuf,
                              MountListBufSize))!=RETURN_OK)
    {
      UBYTE errmsg[FAULT_MAX + strlen(name)];
      IPTR array[] = {filename, errmsg};

      DEBUG_MOUNT(Printf("ReadMountFile: parsemountfile error %ld\n",
                         error));

      #if 1
      switch (error)
      {
        case ERR_DEVICENOTFOUND:
          NewRawDoFmt("Device '%s:' not found",  NULL, errmsg, name);
          break;
        case ERR_INVALIDKEYWORD:
          strcpy(errmsg, "Invalid keyword");
          break;
        default:
          Fault(error, NULL, errmsg, sizeof(errmsg) - 1);
          break;
      }

      displayerror("Mountfile '%s': %s", array);
      #else
      Printf("Mountfile '%s' is invalid: ",filename);
      PrintFault(error, "Mount");
      #endif
    }
    else
    {
      mountinfo	=	TRUE;
    }
    FreeVec(MountListBuf);
  }
  else
  {
    DEBUG_MOUNT(Printf("ReadMountFile: mountfile not found..search for <%s.info>\n",
                       filename));
  }

  if (error == RETURN_OK ||
      error == ERROR_OBJECT_NOT_FOUND)
  {
    DEBUG_MOUNT(Printf("ReadMountFile: look for icon\n"));

    if ((IconBase =  OpenLibrary("icon.library", 37)))
    {
      if ((diskobj = GetDiskObject(filename)))
      {
        myargv	=(char**) diskobj->do_ToolTypes;
        if (myargv)
        {
          while (*myargv)
          {
            char	*ToolPtr;
            ToolPtr	=	*myargv;
            DEBUG_MOUNT(Printf("ReadMountFile: ToolType <%s>\n",
                               ToolPtr));
            if (ToolPtr[0] != '(' && ToolPtr[0] != '*' &&
                !(ToolPtr[0] == 'I' && ToolPtr[1] == 'M' && ToolPtr[3] == '='))
            {
              char	*ToolString;
              toollen	=	strlen(ToolPtr);
              if ((ToolString = AllocVec(toollen + 2,MEMF_ANY)))
              {
                memcpy(ToolString,ToolPtr,toollen);
                ToolString[toollen]	=	'\n';
                ToolString[toollen+1]	=	'\0';
                memclr(&rda,sizeof(struct RDArgs));
                rda.RDA_Source.CS_Buffer	=	ToolString;
                rda.RDA_Source.CS_Length	=	toollen+1;
                rda.RDA_Source.CS_CurChr	=	0;
                rda.RDA_Flags = RDAF_NOPROMPT;
                if ((ReadMountArgs(params,
                                   &rda)==RETURN_OK))
                {
                  mountinfo	=	TRUE;
                }
                else
                {
                  DEBUG_MOUNT(Printf("ReadMountFile: ReadArgs failed error %ld\n",
                                     error));
                }
                FreeVec(ToolString);
              }
              else
              {
                error = ERROR_NO_FREE_STORE;
                break;
              }
            }
            else
            {
              DEBUG_MOUNT(Printf("ReadMountFile: skipped\n"));
            }
            myargv++;
          }
        }
        FreeDiskObject(diskobj);
      }
      else
      {
      }
      CloseLibrary(IconBase);
    }
  }

  if (mountinfo)
  {
    DEBUG_MOUNT(Printf("ReadMountFile: mount information exist\n"));

    if ((error = mount(params,name)) != RETURN_OK)
    {
      DEBUG_MOUNT(Printf("ReadMountFile: mount failed error %ld\n",
                         error));

      if (error == ERROR_OBJECT_EXISTS)
      {
        IPTR array[] = {name};
        displayerror("Device '%s:' is already mounted", array);
      }
    }
  }

  FreeStuff();

  return error;
}

/************************************************************************************************/
/************************************************************************************************/


LONG readfile(STRPTR name, STRPTR *mem, LONG *sizep)
{
    BPTR ml;
    ULONG rest,sub;
    STRPTR buf;
    struct Process *me = (struct Process *) FindTask(NULL);
    APTR oldwinptr;

    oldwinptr = me->pr_WindowPtr;
    me->pr_WindowPtr = (APTR) -1;
    ml = Open(name, MODE_OLDFILE);
    me->pr_WindowPtr = oldwinptr;

    DEBUG_MOUNT(Printf("ReadFile: <%s>\n", (LONG) name));

    if (ml)
    {
	if (Seek(ml, 0, OFFSET_END) != -1)
	{
	    LONG size = Seek(ml, 0, OFFSET_BEGINNING);

	    if (size != -1)
	    {
		*mem = (STRPTR)AllocVec(size + 2, MEMF_ANY);

		if (*mem)
		{
		    rest = size;
		    buf = *mem;

		    for (;;)
		    {
			if (!rest)
			{
			    Close(ml);

			    /* if there is no final linefeed in the buffer, add it */
			    if (buf[-1] != '\n')
			    {
				*buf++ = '\n';
			    	size++;
			    }

			    *buf = '\0';

			    *sizep = size;

			    return 0;
			}

			sub = Read(ml, buf, rest);

			if (sub == -1)
			{
			    break;
			}

			rest -= sub;
			buf += sub;
		    }

		    FreeVec(*mem);
		}
		else
		{
		    SetIoErr(ERROR_NO_FREE_STORE);
		}
	    }
	}

	Close(ml);
    }

    DEBUG_MOUNT(Printf("ReadFile: error %ld\n", IoErr()));
    return IoErr();
}

/************************************************************************************************/
/************************************************************************************************/


void preparefile(STRPTR buf, LONG size)
{
    STRPTR end = buf + size;

    while (buf < end)
    {
	/* Convert comments to spaces */
	if (buf + 1 < end && *buf == '/' && buf[1] == '*')
	{
	    *buf++ = ' ';
	    *buf++ = ' ';

	    while (buf < end)
	    {
		if (*buf == '*')
		{
		    *buf++ = ' ';

		    if (buf >= end)
		    {
			break;
		    }

		    if (*buf == '/')
		    {
			*buf++ = ' ';
			break;
		    }
		}
		else
		{
		    *buf++=' ';
		}
	    }

	    continue;
	}
	
	/* Skip strings */
	if (*buf=='\"')
	{
	    /*
	     * skip first
	     */
	    buf++;
	    while (buf < end && *buf != '\"')
	    {
		buf++;
	    }
	    /*
	     * skip last "
	     */
	    buf++;
	    continue;
	}

	/* Convert '\n' and ';' to spaces */
	if (*buf == '\n' || *buf == ';')
	{
	    *buf++ = ' ';
	    continue;
	}

	/* Convert '#' to NUL */
	if (*buf == '#')
	{
#if 1
/* readargs needs a \n ending */
	    *buf++ = '\n';
#else
	    *buf++ = 0;
#endif
	    continue;
	}

	/* Skip all other characters */
	buf++;
    }

    /* Check the last char, if not newline, force it */
    if (size && end[-1] != '\n')
    {
	end[-1] = '\n';
    }
}

/************************************************************************************************/
/************************************************************************************************/


struct FileSysEntry	*GetFileSysEntry(ULONG	DosType)
{
struct FileSysResource	*MyFileSysRes;
struct FileSysEntry	*MyFileSysEntry;
struct FileSysEntry	*CurrentFileSysEntry;

  MyFileSysEntry		=	NULL;
  if ((MyFileSysRes=OpenResource(FSRNAME)))
  {
    Forbid();
    CurrentFileSysEntry		=(struct FileSysEntry*) MyFileSysRes->fsr_FileSysEntries.lh_Head;
    while (CurrentFileSysEntry->fse_Node.ln_Succ)
    {
      if (CurrentFileSysEntry->fse_DosType == DosType)
      {
        if (MyFileSysEntry)
        {
          if (CurrentFileSysEntry->fse_Version > MyFileSysEntry->fse_Version)
          {
            MyFileSysEntry	=	CurrentFileSysEntry;
          }
        }
        else
        {
          MyFileSysEntry	=	CurrentFileSysEntry;
        }
      }
      CurrentFileSysEntry	=(struct FileSysEntry*) CurrentFileSysEntry->fse_Node.ln_Succ;
    }
    Permit();
  }
  return MyFileSysEntry;
}

/************************************************************************************************/
/************************************************************************************************/


void	PatchDosNode(struct DeviceNode	*MyDeviceNode,
                     ULONG		DosType)
{
struct FileSysEntry	*MyFileSysEntry;
ULONG			MyPatchFlags;

  DEBUG_PATCHDOSNODE(Printf("PatchDosNode: DeviceNode 0x%lx\n",
                    (ULONG)MyDeviceNode));

  if ((MyFileSysEntry=GetFileSysEntry(DosType)))
  {
    MyPatchFlags		=	MyFileSysEntry->fse_PatchFlags;

    DEBUG_PATCHDOSNODE(Printf("PatchDosNode: FileSysEntry 0x%lx PatchFlags 0x%lx\n",
                      (ULONG)MyFileSysEntry,
                      MyPatchFlags));

    if (MyPatchFlags)
    {
      ULONG	*PatchDeviceNode;
      ULONG	*PatchDeviceNodeEnd;
      ULONG	*PatchFileSysEntry;

      PatchFileSysEntry		=(ULONG*) &MyFileSysEntry->fse_Type;
      PatchDeviceNode		=(ULONG*) &MyDeviceNode->dn_Type;
      PatchDeviceNodeEnd	=(ULONG*) ((ULONG) MyDeviceNode + sizeof(struct DeviceNode));

      while (MyPatchFlags)
      {
        if (MyPatchFlags & 1)
        {
          *PatchDeviceNode	=	*PatchFileSysEntry;
        }
        PatchDeviceNode++;
        PatchFileSysEntry++;

        if (PatchDeviceNode >= PatchDeviceNodeEnd)
        {
          /* Savety */
          break;
        }

        MyPatchFlags	>>=	1;
      }
    }
  }
  else
  {
    /* Mount with no BootNode */
    DEBUG_PATCHDOSNODE(Printf("PatchDosNode: Can't get FileSysEntry..no bootnode\n"));
  }
}


/************************************************************************************************/
/************************************************************************************************/



//#define	DOSNAME_INDEX		0
#define	EXECNAME_INDEX		1
#define	UNIT_INDEX		2
#define	FLAGS_INDEX		3
#define	ENVIROMENT_INDEX	4

/* If the underlying device can report the desired access mask and maxtransfer
 * values then override the mountlist provided parameters. - Piru
 */
static void FixParmsFromDeviceQuery(ULONG *ParameterPkt)
{
  struct MsgPort *port;

  if (!ParameterPkt || !ParameterPkt[EXECNAME_INDEX])
    return;

  port = CreateMsgPort();
  if (port)
  {
    struct IOExtTD *ioreq = CreateIORequest(port, sizeof(*ioreq));
    if (ioreq)
    {
      struct Library *QueryBase;
      APTR oldwinptr = MyProcess->pr_WindowPtr;
      MyProcess->pr_WindowPtr = (APTR) -1;

      QueryBase = OpenLibrary("query.library", 51);
      if (QueryBase)
      {
        if (OpenDevice((STRPTR) ParameterPkt[EXECNAME_INDEX], ParameterPkt[UNIT_INDEX], (APTR) ioreq,
                       ParameterPkt[FLAGS_INDEX]) == 0)
        {
          ULONG value;

          if (ParameterPkt[ENVIROMENT_INDEX + DE_TABLESIZE] >= DE_MASK &&
              QueryGetDeviceAttr(ioreq->iotd_Req.io_Device, (ULONG *) &value, QUERYINFOATTR_DEVICE_MASK))
          {
            ParameterPkt[ENVIROMENT_INDEX + DE_MASK] = value;
          }

          if (ParameterPkt[ENVIROMENT_INDEX + DE_TABLESIZE] >= DE_MAXTRANSFER &&
              QueryGetDeviceAttr(ioreq->iotd_Req.io_Device, (ULONG *) &value, QUERYINFOATTR_DEVICE_MAXTRANSFER))
          {
            ParameterPkt[ENVIROMENT_INDEX + DE_MAXTRANSFER] = value;
          }

          CloseDevice((APTR) ioreq);
        }

        CloseLibrary(QueryBase);
      }

      MyProcess->pr_WindowPtr = oldwinptr;

      DeleteIORequest((APTR) ioreq);
    }

    DeleteMsgPort(port);
  }
}

struct DeviceNode	*MyMakeDosNode(char	*DosName,
                                       ULONG	*ParameterPkt,
                                       char	*StartupName)
{
int				DosNameSize;
int				ExecNameSize=0;
int				StartupNameSize=0;
int				MyEnvSize=0;
struct DeviceNode		*MyDeviceNode=NULL;
struct FileSysStartupMsg	*MyFileSysStartupMsg=NULL;
struct DosEnvec			*MyDosEnvec=NULL;
char				*MyString=NULL;
ULONG				Status=FALSE;
  DEBUG_MAKEDOSNODE(Printf("MakeDosNode: Pkt 0x%lx\n",
                            (ULONG)ParameterPkt));
  
  if (ParameterPkt)
  {
    DEBUG_MAKEDOSNODE(Printf("MakeDosNode: DosName <%s> DeviceName <%s> Unit %ld Flags 0x%lx DE_TABLESIZE 0x%lx\n",
                             (ULONG)DosName,
                             ParameterPkt[EXECNAME_INDEX],
                             ParameterPkt[UNIT_INDEX],
                             ParameterPkt[FLAGS_INDEX],
                             ParameterPkt[ENVIROMENT_INDEX]));
  }
  else
  {
    DEBUG_MAKEDOSNODE(Printf("MakeDosNode: DosName <%s> Startup <%s>\n",
                             (ULONG)DosName,
                             (ULONG)StartupName));
  }

  DosNameSize	=	strlen(DosName);

  FixParmsFromDeviceQuery(ParameterPkt);

  if (ParameterPkt)
  {
    if (ParameterPkt[EXECNAME_INDEX])
    {
      ExecNameSize	=	strlen((UBYTE*) ParameterPkt[EXECNAME_INDEX]);
    }
    else
    {
      ExecNameSize	=	0;
    }
    MyEnvSize		=	(ParameterPkt[ENVIROMENT_INDEX] + 1) * sizeof(ULONG);
  }
  else
  {
    StartupNameSize	=	StartupName ? strlen(StartupName) : 0;
  }

  if ((MyDeviceNode=AllocVec(sizeof(struct DeviceNode),
                             MEMF_PUBLIC | MEMF_CLEAR)))
  {
    DEBUG_MAKEDOSNODE(Printf("MakeDosNode: MyDeviceNode 0x%lx\n",
                              (ULONG)MyDeviceNode));

    MyDeviceNode->dn_StackSize		=	600;
    MyDeviceNode->dn_Priority		=	10;

    if ((MyString=AllocVec(((DosNameSize+2+4) & ~3) +
                           ((ExecNameSize+2+4) & ~3) +
                           ((StartupNameSize+2+2+4) & ~3),
                           MEMF_PUBLIC | MEMF_CLEAR)))
    {
      MyString[0]		=	DosNameSize;

      strcpy(&MyString[1],
             DosName);

      MyDeviceNode->dn_Name	=	MKBADDR(MyString);


      if (ParameterPkt)
      {
        if ((MyFileSysStartupMsg=AllocVec(sizeof(struct FileSysStartupMsg),
                                          MEMF_PUBLIC | MEMF_CLEAR)))
        {
          DEBUG_MAKEDOSNODE(Printf("MakeDosNode: MyFileSysStartupMsg 0x%lx\n",
                                    (ULONG)MyFileSysStartupMsg));

          if ((MyDosEnvec=AllocVec(MyEnvSize,
                                   MEMF_PUBLIC | MEMF_CLEAR)))
          {
            char	*ExecNamePtr;

            DEBUG_MAKEDOSNODE(Printf("MakeDosNode: MyDosEnvec 0x%lx\n",
                                      (ULONG)MyDosEnvec));


            #if 1

            ExecNamePtr		=	&MyString[(1+DosNameSize+2+3) & ~3];

            /* .device name must absolutely **NOT** include the 0 in the
             * length!!
             *
             * the string *MUST* be 0 terminated, however!
             */
            ExecNamePtr[0]	=	ExecNameSize;

            if (ParameterPkt[EXECNAME_INDEX])
            {
              strcpy(&ExecNamePtr[1],
                     (UBYTE*) ParameterPkt[EXECNAME_INDEX]);
            }

            MyFileSysStartupMsg->fssm_Device	=	MKBADDR(ExecNamePtr);

            #else

            /* IMO this is more correct than above, but I could be wrong.. - Piru */

            if (ParameterPkt[EXECNAME_INDEX])
            {
              ExecNamePtr		=	&MyString[(1+DosNameSize+2+3) & ~3];

              /* .device name must absolutely **NOT** include the 0 in the
               * length!!
               *
               * the string *MUST* be 0 terminated, however!
               */
              ExecNamePtr[0]	=	ExecNameSize;

              strcpy(&ExecNamePtr[1],
                     (UBYTE*) ParameterPkt[EXECNAME_INDEX]);

              MyFileSysStartupMsg->fssm_Device	=	MKBADDR(ExecNamePtr);
            }
            #endif

            MyFileSysStartupMsg->fssm_Unit	=	ParameterPkt[UNIT_INDEX];
            MyFileSysStartupMsg->fssm_Flags	=	ParameterPkt[FLAGS_INDEX];
            MyFileSysStartupMsg->fssm_Environ	=	MKBADDR(MyDosEnvec);
            MyDeviceNode->dn_Startup		=	MKBADDR(MyFileSysStartupMsg);

            CopyMem(&ParameterPkt[ENVIROMENT_INDEX],
                    MyDosEnvec,
                    MyEnvSize);

            Status=TRUE;
            //DEBUG_MAKEDOSNODE(Printf("MakeDosNode: done\n"));
          }
        }
      }
      else
      {
        if (StartupName && StartupNameSize)
        {
          char	*StartupNamePtr;
#if 0
          BOOL Space=FALSE;
          int i;

/*
 * This code can't be normal..disabled
 * Ralph 10.11.2003
 */
          for (i=0;i<StartupNameSize;i++)
          {
            if (StartupName[i]==' ')
            {
              DEBUG_MAKEDOSNODE(Printf("MakeDosNode: SPACE found\n"));
              Space = TRUE;
              break;
            }
          }
          if (Space)
          {
            StartupNamePtr	=	&MyString[(1+DosNameSize+2+3) & ~3];
            StartupNamePtr[0]	=	StartupNameSize+2;
            StartupNamePtr[1]	=	'"';

            strcpy(&StartupNamePtr[2],
                   StartupName);
            StartupNamePtr[2+StartupNameSize]	=	'"';
            StartupNamePtr[2+StartupNameSize+1]	=	'\0';
          }
          else
#endif
          {
            StartupNamePtr	=	&MyString[(1+DosNameSize+2+3) & ~3];
            StartupNamePtr[0]	=	StartupNameSize;

            strcpy(&StartupNamePtr[1],
                   StartupName);
          }
          MyDeviceNode->dn_Startup	=	MKBADDR(StartupNamePtr);
        }
        Status=TRUE;
      }
    }
  }
  if (Status)
  {
    DEBUG_MAKEDOSNODE(Printf("MakeDosNode: done\n"));
    return MyDeviceNode;
  }
  else
  {
    DEBUG_MAKEDOSNODE(Printf("MakeDosNode: failed\n"));
    FreeVec(MyDosEnvec);
    FreeVec(MyFileSysStartupMsg);
    FreeVec(MyDeviceNode);
    FreeVec(MyString);
    return NULL;
  }
}



/************************************************************************************************/
/************************************************************************************************/


LONG	parsemountfile(IPTR		*params,
                       STRPTR		buf,
                       LONG		size)
{
struct Args	args;
LONG   error;
struct RDArgs rda;

    DEBUG_MOUNT(Printf("ParseMountFile:\n"));

    memclr(&args,sizeof(struct Args));
    memclr(&rda,sizeof(struct RDArgs));

    rda.RDA_Source.CS_Buffer = buf;
    rda.RDA_Source.CS_Length = size;
    rda.RDA_Source.CS_CurChr = 0;
    rda.RDA_Flags = RDAF_NOPROMPT;

    DEBUG_MOUNT(Printf("ReadArgs..\n%s\n\n",(ULONG)rda.RDA_Source.CS_Buffer));

    if ((error=ReadMountArgs(params,
                            &rda))!=RETURN_OK)
    {
	DEBUG_MOUNT(Printf("Parse: ReadArgs failed\n"));
    }
    return error;
}

/************************************************************************************************/
/************************************************************************************************/


LONG	parsemountlist(IPTR		*params,
                       STRPTR		name,
                       STRPTR		buf,
                       LONG		size)
{
struct Args	args;
UBYTE  buffer[1024];
LONG   error=RETURN_OK, res;
STRPTR end = buf + size;
STRPTR s2;
char *ptr;
struct RDArgs rda;

    DEBUG_MOUNT(Printf("ParseMountList: <%s>\n",(ULONG)name));

    memclr(&args,sizeof(struct Args));
    memclr(&rda,sizeof(struct RDArgs));

    rda.RDA_Source.CS_Buffer = buf;
    rda.RDA_Source.CS_Length = end - buf;
    rda.RDA_Source.CS_CurChr = 0;
    rda.RDA_Flags = RDAF_NOPROMPT;

    while (rda.RDA_Source.CS_CurChr < rda.RDA_Source.CS_Length)
    {
	res = ReadItem(buffer,
                       sizeof(buffer),
                       &rda.RDA_Source);

	DEBUG_MOUNT(Printf("ParseMountList: buffer <%s>\n",(LONG) (res > ITEM_NOTHING ? (char *) buffer : (char *) "")));
	DEBUG_MOUNT(Printf("ParseMountList: ReadItem res %ld\n",res));

	if (res == ITEM_ERROR)
	{
	    return IoErr();
	}

	#if 0
	if (res == ITEM_NOTHING &&
	    rda.RDA_Source.CS_CurChr == rda.RDA_Source.CS_Length)
	{
	    return 0;
	}

	if (res != ITEM_QUOTED && res != ITEM_UNQUOTED)
	{
	    return 1;
	}
	#endif

	s2 = buffer;

	while (*s2)
	{
	    s2++;
	}

	if (s2 == buffer || s2[-1] != ':')
	{
	    goto skipnext;
	    //DEBUG_MOUNT(Printf("ParseMountList: failure\n"));
	    //return ERR_DEVICENOTFOUND;
	}

	*--s2 = '\0';

	if (!Strnicmp(name, buffer, s2 - buffer) &&
	   (!name[s2 - buffer] || (name[s2 - buffer] == ':' || !name[s2 - buffer + 1])))
	{
	    DEBUG_MOUNT(Printf("ParseMountList: found\n"));

	    /* Copy the string so we get proper case - Piru */
	    memcpy(name, buffer, s2 - buffer);
	    name[s2 - buffer] = '\0';

	    ptr = name;
	    while (*ptr)
	    {
		if (*ptr++ == ':')
		{
			ptr[-1] = '\0';
			break;
		}
	    }

	    DEBUG_MOUNT(Printf("ReadArgs..\n%s\n\n",(ULONG)&rda.RDA_Source.CS_Buffer[rda.RDA_Source.CS_CurChr]));

	    if ((error=ReadMountArgs(params,
                                     &rda))!=RETURN_OK)
	    {
		DEBUG_MOUNT(Printf("ParseMountList: ReadArgs failed\n"));
		//return IoErr();
	    }

	    return error;
	}

	skipnext:
	while (rda.RDA_Source.CS_CurChr < rda.RDA_Source.CS_Length)
	{
	    if (rda.RDA_Source.CS_Buffer[rda.RDA_Source.CS_CurChr++] == '\n')
	    {
		DEBUG_MOUNT(Printf("ParseMountList: reach the end of the block\n"));
		break;
	    }
	}
    }

    DEBUG_MOUNT(Printf("ParseMountList: mount found nothing\n"));
    return ERR_DEVICENOTFOUND;
}

/************************************************************************************************/
/************************************************************************************************/

LONG	checkmount(IPTR		*params)
{
struct DosEnvec *vec;

	vec = (struct DosEnvec *)&params[4];

	/* bootpri -129 shouldn't be started and not automatic mounted..whatever that means */
	if (vec->de_BootPri < -129 || vec->de_BootPri > 127)
	{
		if (MyProcess->pr_CLI)
			Printf("CheckMount: BootPri %ld is not allowed. Legal range is -128..127\n",vec->de_BootPri);
		return ERROR_BAD_NUMBER;
	}

	if (flagargs.GlobalVec)
	{
		if (GlobalVec < -3 || GlobalVec > -1)
		{
			if (MyProcess->pr_CLI)
				Printf("CheckMount: Globvec %ld is not supported. Only -3, -2 and -1 are supported here\n",
        			       GlobalVec);
			return ERROR_BAD_NUMBER;
		}
	}

	if (flagargs.Startup && !StartupString)
	{
		if (StartupValue >= 0x100)
		{
			if (MyProcess->pr_CLI)
				Printf("CheckMount: Startup uses a too large value %ld, maximum is 255\n",StartupValue);
			return ERROR_BAD_NUMBER;
		}
	}

	return RETURN_OK;
}

/************************************************************************************************/
/************************************************************************************************/

LONG mount(IPTR		*params,
           STRPTR	name)
                  
{
struct DosEnvec *vec;
LONG		error = RETURN_OK;
ULONG		Flags;

    DEBUG_MOUNT(Printf("MountDev: <%s>\n",(ULONG)name));

    if ((error=checkmount(params))!=RETURN_OK)
    {
	DEBUG_MOUNT(Printf("MountDev: checkmount failed\n"));
	return error;
    }

    vec = (struct DosEnvec *)&params[4];

#if 0
    if (!HandlerString && !flagargs.DosType)
    {
	Printf("Mount: No specified Handler or DosType\n");
	return ERROR_REQUIRED_ARG_MISSING;
    }
#endif

    DEBUG_MOUNT(Printf("MountDev: DosName <%s>\n",(ULONG)name));
    DEBUG_MOUNT(Printf("MountDev: Filesystem <%s>\n",(ULONG)(HandlerString ? (HandlerString+1) : "")));
    DEBUG_MOUNT(Printf("MountDev: Device  <%s>\n",(ULONG)DeviceString));

    DEBUG_MOUNT(Printf("MountDev:         TableSize %ld\n",vec->de_TableSize));
    DEBUG_MOUNT(Printf("MountDev:         SizeBlock %ld\n",vec->de_SizeBlock));
    DEBUG_MOUNT(Printf("MountDev:            SecOrg %ld\n",vec->de_SecOrg));
    DEBUG_MOUNT(Printf("MountDev:          Surfaces %ld\n",vec->de_Surfaces));
    DEBUG_MOUNT(Printf("MountDev:   SectorsPerBlock %ld\n",vec->de_SectorPerBlock));
    DEBUG_MOUNT(Printf("MountDev:    BlocksPerTrack %ld\n",vec->de_BlocksPerTrack));
    DEBUG_MOUNT(Printf("MountDev:          Reserved %ld\n",vec->de_Reserved));
    DEBUG_MOUNT(Printf("MountDev:          PreAlloc %ld\n",vec->de_PreAlloc));
    DEBUG_MOUNT(Printf("MountDev:        Interleave %ld\n",vec->de_Interleave));
    DEBUG_MOUNT(Printf("MountDev:            LowCyl %lu\n",vec->de_LowCyl));
    DEBUG_MOUNT(Printf("MountDev:          UpperCyl %lu\n",vec->de_HighCyl));
    DEBUG_MOUNT(Printf("MountDev:        NumBuffers %ld\n",vec->de_NumBuffers));
    DEBUG_MOUNT(if (vec->de_TableSize >= DE_BUFMEMTYPE))
     DEBUG_MOUNT(Printf("MountDev:        BufMemType 0x%lx\n",vec->de_BufMemType));
    DEBUG_MOUNT(Printf("MountDev:       MaxTransfer 0x%lx\n",vec->de_MaxTransfer));
    DEBUG_MOUNT(if (vec->de_TableSize >= DE_MASK))
     DEBUG_MOUNT(Printf("MountDev:              Mask 0x%lx\n",vec->de_Mask));
    DEBUG_MOUNT(if (vec->de_TableSize >= DE_BOOTPRI))
     DEBUG_MOUNT(Printf("MountDev:           BootPri %ld\n",vec->de_BootPri));
    DEBUG_MOUNT(if (vec->de_TableSize >= DE_DOSTYPE))
     DEBUG_MOUNT(Printf("MountDev:           DosType 0x%lx\n",vec->de_DosType));
    DEBUG_MOUNT(if (vec->de_TableSize >= DE_BAUD))
     DEBUG_MOUNT(Printf("MountDev:              Baud %ld\n",vec->de_Baud));
    DEBUG_MOUNT(if (vec->de_TableSize >= DE_CONTROL))
     DEBUG_MOUNT(Printf("MountDev:           Control 0x%lx\n",vec->de_Control));
    DEBUG_MOUNT(if (vec->de_TableSize >= DE_BOOTBLOCKS))
     DEBUG_MOUNT(Printf("MountDev:        BootBlocks %ld\n",vec->de_BootBlocks));

    {
	struct DeviceNode *dn;

	/* If Device is specified, do create FSSM */
	if (NoFSSM && DeviceString)
	{
	    NoFSSM = FALSE;
	}

	params[1] = (IPTR) DeviceString;

	if ((dn=MyMakeDosNode(name,
                              (flagargs.Startup || NoFSSM) ? NULL : params,
                              StartupString)))
	{
	    DEBUG_MOUNT(Printf("MountDev: DeviceNode 0x%lx\n",(ULONG)dn));


	    dn->dn_StackSize	=	StackSize;
	    dn->dn_Priority	=	Priority;
	    dn->dn_GlobalVec	=	GlobalVec;

	    if (flagargs.Startup && !StartupString)
	    {
		dn->dn_Startup = StartupValue;
	    }

	    if (!ForceLoad && !HandlerString)
	    {
		DEBUG_MOUNT(Printf("MountDev: patchdosnode\n"));
		PatchDosNode(dn,vec->de_DosType);
	    }

	    if (ForceLoad || dn->dn_SegList==NULL)
	    {
		DEBUG_MOUNT(Printf("MountDev: Load Handler\n"));
		dn->dn_Handler	=	MKBADDR(HandlerString);
	    }
	    else
	    {
		/*
		 * We don't need the HandlerString anymore...free it
		 */
		if (HandlerString)
		{
		    FreeVec(HandlerString);
		    HandlerString	=	NULL;
		}
	    }
	    DEBUG_MOUNT(Printf("MountDev:      Name %b\n",dn->dn_Name));
	    DEBUG_MOUNT(Printf("MountDev:   Handler 0x%lx <%b>\n",dn->dn_Handler,dn->dn_Handler));
	    DEBUG_MOUNT(Printf("MountDev:   SegList 0x%lx\n",dn->dn_SegList));
	    DEBUG_MOUNT(Printf("MountDev: StackSize %ld\n",dn->dn_StackSize));
	    DEBUG_MOUNT(Printf("MountDev:  Priority %ld\n",dn->dn_Priority));
	    DEBUG_MOUNT(Printf("MountDev:   Startup 0x%lx\n",dn->dn_Startup));
	    DEBUG_MOUNT(Printf("MountDev: GlobalVec %ld\n",dn->dn_GlobalVec));

	    if (dn->dn_SegList || dn->dn_Handler)
	    {
		if (Activate)
		{
			Flags = ADNF_STARTPROC;
		}
		else
		{
			Flags = 0;
		}
//__builtin_emit(0x4afc);
		if (CheckDevice(name) == 0)
		{
			if (AddDosNode(vec->de_BootPri, Flags , dn))
			{
				DEBUG_MOUNT(Printf("MountDev: AddDosNode worked\n"));
				/*
				 * Don't free these anymore as they belong to the dosnode
				 */
                                UnitString	=	NULL;
                                FlagsString	=	NULL;
	                        HandlerString	=	NULL;
                                ControlString	=	NULL;
				error = 0;
			}
			else
			{
				DEBUG_MOUNT(Printf("MountDev: AddDosNode failed\n"));
				error = ERROR_INVALID_RESIDENT_LIBRARY;
				#if 0
				if (HandlerString)
				{
					FreeVec(HandlerString);
				}
				#endif
			}
		}
		else
		{
			DEBUG_MOUNT(Printf("MountDev: Device is already mounted\n"));
			error = ERROR_OBJECT_EXISTS;
		}
	    }
	    else
	    {
		    DEBUG_MOUNT(Printf("MountDev: no loadseg and no handler specified\n"));
		    error = ERROR_OBJECT_NOT_FOUND;
	    }
	}
	else
	{
	    error = ERROR_NO_FREE_STORE;
	}
    }

    return error;
}
